/*
 * This file is part of EasyRPG Player.
 *
 * EasyRPG Player is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EasyRPG Player is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EasyRPG Player. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef EP_SCENE_H
#define EP_SCENE_H

// Headers
#include "graphics.h"
#include "system.h"
#include <vector>

/**
 * Scene virtual class.
 */
class Scene {
public:
	/** Scene types. */
	enum SceneType {
		Null,
		Title,
		Map,
		Menu,
		Item,
		Skill,
		Equip,
		ActorTarget,
		Status,
		File,
		Save,
		Load,
		End,
		Battle,
		Shop,
		Name,
		Gameover,
		Debug,
		Logo,
		Order,
		GameBrowser,
		Teleport,
		SceneMax
	};

	/**
	 * Constructor.
	 */
	Scene();

	/**
	 * Destructor.
	 */
	virtual ~Scene() {};

	/**
	 * Scene entry point.
	 * The Scene Main-Function manages a stack and always
	 * executes the scene that is currently on the top of
	 * the stack.
	 */
	virtual void MainFunction();

	/**
	 * Start processing.
	 * This function is executed while the screen is faded
	 * out. All objects needed for the scene should be
	 * created here.
	 */
	virtual void Start();

	/**
	 * Continue processing.
	 * This function is executed when returning from a
	 * nested scene (instead of Start).
	 *
	 * @param prev_scene The previous scene
	 */
	virtual void Continue(SceneType prev_scene);

	/**
	 * Resume processing.
	 * This function is executed after the fade in,
	 * either when starting the scene or when returning
	 * from a nested scene
	 *
	 * @param prev_scene The previous scene
	 */
	virtual void Resume(SceneType prev_scene);

	/**
	 * Suspend processing.
	 * This function is executed before the fade out for
	 * the scene change, either when terminating the scene
	 * or switching to a nested scene
	 */
	virtual void Suspend();

	/**
	 * Does the transition upon starting or resuming
	 * the scene
	 *
	 * @param prev_scene the scene we transitioned from
	 */
	virtual void TransitionIn(SceneType prev_scene);

	/**
	 * Does the transition upon ending or suspending
	 * the scene
	 *
	 * @param next_scene the scene we will transition to
	 */
	virtual void TransitionOut(SceneType next_scene);

	/**
	 * Called when a transition or async load is finished.
	 */
	void OnFinishAsync();

	/**
	 * Tell the scene we spawned an async operation from the main loop.
	 * This is used to delay incrementing the frame counter until all
	 * async operations and continuations are completed.
	 */
	void SetAsyncFromMainLoop();

	/** @returns true if an async operation that would block the main loop is pending */
	static bool IsAsyncPending();

	/**
	 * Called every frame.
	 * The scene should redraw all elements.
	 */
	virtual void Update();

	/**
	 * Pushes a new scene on the scene execution stack.
	 *
	 * @param new_scene new scene.
	 * @param pop_stack_top if the scene that is currently
	 *                      on the top should be popped.
	 */
	static void Push(std::shared_ptr<Scene> const& new_scene, bool pop_stack_top = false);

	/**
	 * Removes the scene that is on the top of the stack.
	 */
	static void Pop();

	/**
	 * Removes scenes from the stack, until a specific one
	 * is reached.
	 *
	 * @param type type of the scene that is searched.
	 */
	static void PopUntil(SceneType type);

	/**
	 * Finds the topmost scene of a specific type on the stack.
	 *
	 * @param type type of the scene that is searched.
	 * @return the scene found, or NULL if no such scene exists.
	 */
	static std::shared_ptr<Scene> Find(SceneType type);

	// Don't write to the following values directly when you want to change
	// the scene! Use Push and Pop instead!

	/** Scene type. */
	SceneType type;

	/** Current scene. */
	static std::shared_ptr<Scene> instance;

	/** Old scenes, temporary save for deleting. */
	static std::vector<std::shared_ptr<Scene> > old_instances;

	/** Contains name of the Scenes. For debug purposes. */
	static const char scene_names[SceneMax][12];

	/** Called by the graphic system to request drawing of a background, usually a system color background */
	virtual void DrawBackground();

	Graphics::State& GetGraphicsState();

	/** Called by the interpreter when an event finishes processing */
	virtual void onCommandEnd() {}

	/** @return true if the Scene has been initialized */
	bool IsInitialized() const;

	/** @return the scene requested by events */
	SceneType GetRequestedScene();

	/** @return true if a scene is being requested */
	bool HasRequestedScene();

	/**
	 * Sets the requested scene to change to
	 *
	 * @param scene the scene to call
	 */
	void SetRequestedScene(SceneType scene);

protected:
	using AsyncContinuation = std::function<void(void)>;
	AsyncContinuation async_continuation;

private:
	/** Scene stack. */
	static std::vector<std::shared_ptr<Scene> > instances;

	static int push_pop_operation;

	/**
	 * true if Start() was called. For handling the special case that two
	 * or more scenes are pushed. In that case only the last calls start, the
	 * other Continue(). This enforces calling Start().
	 */
	bool initialized = false;

	bool was_async_from_main_loop = false;

	/**
	 * Graphic stack of the scene
	 */
	Graphics::State state;

	static void DebugValidate(const char* caller);
	static void UpdatePrevScene();

	Scene::SceneType request_scene = Null;
};

inline bool Scene::IsInitialized() const {
	return initialized;
}

inline Scene::SceneType Scene::GetRequestedScene() {
	return request_scene;
}

inline bool Scene::HasRequestedScene() {
	return GetRequestedScene() != Null;
}

inline void Scene::SetRequestedScene(SceneType scene) {
	request_scene = scene;
}


#endif
